#ifndef MAKEDEPEND
#include <cstdio>
#include <cstdarg>
#endif

#include "dasm_expr.h"
#include "disasm.h"

#define DEBUGPRINTS 0

static expr *B,*C,*D,*E, *H,*L,  *A;
static expr *S,*P, *mapper;

#define regname(r) ((&r==&B)?"DH" \
                   :(&r==&C)?"DL" \
                   :(&r==&D)?"CH" \
                   :(&r==&E)?"CL" \
                   :(&r==&H)?"BH" \
                   :(&r==&L)?"BL" \
                   :(&r==&A)?"AL" \
                   :(&r==&S)?"S" \
                   :(&r==&P)?"P" \
                   :"?")
                  
#define regid(r) ((&r==&B)?'B' \
                 :(&r==&C)?'C' \
                 :(&r==&D)?'D' \
                 :(&r==&E)?'E' \
                 :(&r==&H)?'H' \
                 :(&r==&L)?'L' \
                 :(&r==&A)?'A' \
                 :(&r==&S)?'S' \
                 :(&r==&P)?'P' \
                 :0)
                 
#define pairname(r1,r2) \
                 ((&r1==&B)?"DX"\
                 :(&r1==&D)?"CX"\
                 :(&r1==&H)?"BX"\
                 :(&r1==&S)?"SP"\
                 :"??")

#if DEBUGPRINTS
 #define dprintf(x...) printf(x)
#else
 #define dprintf(x...) while(0)
#endif 

const string checkpair(expr *l, expr *r)
{
	if(l==B && r==C)return "DX";
	if(l==D && r==E)return "CX";
	if(l==H && r==L)return "BX";
	if(l==S && r==P)return "SP";
	return "";
}

#ifdef __GNUC__
static void output(const char *fmt, ...)
     __attribute__ ((__format__ (__printf__, 1, 2)));
#endif

static int indentlevel=0;
static bool pendingoutput=false;

void flushpendingoutput()
{
   	if(pendingoutput)
   	{
   		pendingoutput=false;
   		printf("\n");
   	}
}

static void output(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	flushpendingoutput();
	printf("%*s", indentlevel, "");
	pendingoutput=true;
	vprintf(fmt, ap);
	va_end(ap);
}

static void reset(expr *&r)
{
	if(r && r->regid == regid(r))return;
	if(r)dprintf("Resetting '%s' (contained '%s')\n", regname(r), r->debug().c_str());
    if(r && !--r->refnum) delete r;
    r = new expr(regname(r), regid(r));
    ++r->refnum;
}

static string getval(expr *&r)
{
	if(!r->regid)
	{
		string tmp = r->debug();
		
		if((tmp == regname(B) && r != B)
		|| (tmp == regname(C) && r != C)
		|| (tmp == regname(D) && r != D)
		|| (tmp == regname(E) && r != E)
		|| (tmp == regname(H) && r != H)
		|| (tmp == regname(L) && r != L)
		|| (tmp == regname(S) && r != S)
		|| (tmp == regname(P) && r != P)
		|| (tmp == regname(A) && r != A))
		{
			reset(r);
			/* resets it to its default content */
		}
	}
	
	return r->debug();
}

static void flush(expr *&r, bool resetalso=false)
{
    if(r && !r->regid && !r->notneedflush)
    {
    	r->notneedflush = true;
    	output("%s = %s;", regname(r), getval(r).c_str());
    }
    if(resetalso || (regid(r) && r->depends(regid(r))))
    	reset(r);
}

static void showcontents()
{
	dprintf("// \33[32m'%s' (at %X) contains '%s' (ref %u)\33[37m\n", regname(B), (unsigned)B, getval(B).c_str(), B->refnum);
	dprintf("// \33[32m'%s' (at %X) contains '%s' (ref %u)\33[37m\n", regname(C), (unsigned)C, getval(C).c_str(), C->refnum);
	dprintf("// \33[32m'%s' (at %X) contains '%s' (ref %u)\33[37m\n", regname(D), (unsigned)D, getval(D).c_str(), D->refnum);
	dprintf("// \33[32m'%s' (at %X) contains '%s' (ref %u)\33[37m\n", regname(E), (unsigned)E, getval(E).c_str(), E->refnum);
	dprintf("// \33[32m'%s' (at %X) contains '%s' (ref %u)\33[37m\n", regname(H), (unsigned)H, getval(H).c_str(), H->refnum);
	dprintf("// \33[32m'%s' (at %X) contains '%s' (ref %u)\33[37m\n", regname(L), (unsigned)L, getval(L).c_str(), L->refnum);
	dprintf("// \33[32m'%s' (at %X) contains '%s' (ref %u)\33[37m\n", regname(A), (unsigned)A, getval(A).c_str(), A->refnum);
	dprintf("// \33[32m'%s' (at %X) contains '%s' (ref %u)\33[37m\n", regname(S), (unsigned)S, getval(S).c_str(), S->refnum);
	dprintf("// \33[32m'%s' (at %X) contains '%s' (ref %u)\33[37m\n", regname(P), (unsigned)P, getval(P).c_str(), P->refnum);
}

static void assign(expr *&r, expr *n)
{
	expr *tmp = n->calculate();
	++tmp->refnum;
	
	dprintf("// \33[35massigning '%s' (ref %u) into something\33[37m\n", tmp->debug().c_str(), tmp->refnum);
	dprintf("// \33[35mtarget contains: '%s'\33[37m\n", r->debug().c_str());
	if(B != r && B->depends(regid(r))){dprintf("// \33[35m%s (%s) depends on %s, flushing it\33[37m\n", regname(B), getval(B).c_str(), getval(r).c_str());if(B->debug()==tmp->debug()){dprintf("// \33[35m It will be the new source.\33[37m\n");if(!--tmp->refnum)delete tmp;flush(B,true);tmp=B;++tmp->refnum;}else flush(B,true);}
	if(C != r && C->depends(regid(r))){dprintf("// \33[35m%s (%s) depends on %s, flushing it\33[37m\n", regname(C), getval(C).c_str(), getval(r).c_str());if(C->debug()==tmp->debug()){dprintf("// \33[35m It will be the new source.\33[37m\n");if(!--tmp->refnum)delete tmp;flush(C,true);tmp=C;++tmp->refnum;}else flush(C,true);}
	if(D != r && D->depends(regid(r))){dprintf("// \33[35m%s (%s) depends on %s, flushing it\33[37m\n", regname(D), getval(D).c_str(), getval(r).c_str());if(D->debug()==tmp->debug()){dprintf("// \33[35m It will be the new source.\33[37m\n");if(!--tmp->refnum)delete tmp;flush(D,true);tmp=D;++tmp->refnum;}else flush(E,true);}
	if(E != r && E->depends(regid(r))){dprintf("// \33[35m%s (%s) depends on %s, flushing it\33[37m\n", regname(E), getval(E).c_str(), getval(r).c_str());if(E->debug()==tmp->debug()){dprintf("// \33[35m It will be the new source.\33[37m\n");if(!--tmp->refnum)delete tmp;flush(E,true);tmp=E;++tmp->refnum;}else flush(D,true);}
	if(H != r && H->depends(regid(r))){dprintf("// \33[35m%s (%s) depends on %s, flushing it\33[37m\n", regname(H), getval(H).c_str(), getval(r).c_str());if(H->debug()==tmp->debug()){dprintf("// \33[35m It will be the new source.\33[37m\n");if(!--tmp->refnum)delete tmp;flush(H,true);tmp=H;++tmp->refnum;}else flush(H,true);}
	if(L != r && L->depends(regid(r))){dprintf("// \33[35m%s (%s) depends on %s, flushing it\33[37m\n", regname(L), getval(L).c_str(), getval(r).c_str());if(L->debug()==tmp->debug()){dprintf("// \33[35m It will be the new source.\33[37m\n");if(!--tmp->refnum)delete tmp;flush(L,true);tmp=L;++tmp->refnum;}else flush(L,true);}
	if(A != r && A->depends(regid(r))){dprintf("// \33[35m%s (%s) depends on %s, flushing it\33[37m\n", regname(A), getval(A).c_str(), getval(r).c_str());if(A->debug()==tmp->debug()){dprintf("// \33[35m It will be the new source.\33[37m\n");if(!--tmp->refnum)delete tmp;flush(A,true);tmp=A;++tmp->refnum;}else flush(A,true);}
	if(S != r && S->depends(regid(r))){dprintf("// \33[35m%s (%s) depends on %s, flushing it\33[37m\n", regname(S), getval(S).c_str(), getval(r).c_str());if(S->debug()==tmp->debug()){dprintf("// \33[35m It will be the new source.\33[37m\n");if(!--tmp->refnum)delete tmp;flush(S,true);tmp=S;++tmp->refnum;}else flush(S,true);}
	if(P != r && P->depends(regid(r))){dprintf("// \33[35m%s (%s) depends on %s, flushing it\33[37m\n", regname(P), getval(P).c_str(), getval(r).c_str());if(P->debug()==tmp->debug()){dprintf("// \33[35m It will be the new source.\33[37m\n");if(!--tmp->refnum)delete tmp;flush(P,true);tmp=P;++tmp->refnum;}else flush(P,true);}
	dprintf("// \33[35mnew content: '%s' (ref %u)\33[37m\n", tmp->debug().c_str(), tmp->refnum);
	if(r != tmp)
	{
    	if(r && !--r->refnum)delete r;
	    r = tmp;
    	
    	flushpendingoutput();
		dprintf("// \33[35m%s got '%s'\33[37m\n", regname(r), getval(r).c_str());
		if(r->depends(regid(r)))
		{
			dprintf("// \33[35m%s (%s) depends on itself, flushing it\33[37m\n", regname(r), getval(r).c_str());flush(r,true);
	//		showcontents();
		}			
	}
}
static void DatAssign(expr *adr, expr *&src)
{
	string s = adr->debug();
	string t = src->debug();
//	if(src->depends(s))
	{
		if(B->depends(s)) { flush(B, true); }
		if(C->depends(s)) { flush(C, true); }
		if(D->depends(s)) { flush(D, true); }
		if(E->depends(s)) { flush(E, true); }
		if(H->depends(s)) { flush(H, true); }
		if(L->depends(s)) { flush(L, true); }
		if(A->depends(s)) { flush(A, true); }
		if(S->depends(s)) { flush(S, true); }
		if(P->depends(s)) { flush(P, true); }
	/*
		if(getval(B) == t) { flush(B, true); }
		if(getval(C) == t) { flush(C, true); }
		if(getval(D) == t) { flush(D, true); }
		if(getval(E) == t) { flush(E, true); }
		if(getval(H) == t) { flush(H, true); }
		if(getval(L) == t) { flush(L, true); }
		if(getval(A) == t) { flush(A, true); }
		if(getval(S) == t) { flush(S, true); }
		if(getval(P) == t) { flush(P, true); }
	*/
	}
	output("data(%s) = %s", s.c_str(), getval(src).c_str());
	if(!--adr->refnum) delete adr;
}
static void flushpair(expr *&h, expr *&l, bool resetalso=false)
{
	if(h->isconst && l->isconst
	&& !(h->notneedflush && l->notneedflush))
	{
		unsigned w = (h->getnum() << 8) | l->getnum();
		output("%s = 0x%X;", pairname(h,l), w);
		if(resetalso)
		{
			reset(h);
			reset(l);
		}
		else
		{
			h->notneedflush = true;
			l->notneedflush = true;
		}
	}
	else
	{
		flush(h, resetalso);
		flush(l, resetalso);
	}
}

static void flushall(bool resetalso=false)
{
	flush(A, resetalso);
	flushpair(H,L, resetalso);
	flushpair(D,E, resetalso);
	flushpair(B,C, resetalso);
	flushpair(S,P, resetalso);	
}
    
static void resetall()
{
	reset(B); reset(C); reset(D); reset(E);
	reset(H); reset(L); reset(A);
	reset(S); reset(P);
}

#define assign_word(h, l) do{ \
    assign(h, new expr_num(b2)); \
    assign(l, new expr_num(b1)); \
    data+=2; }while(0)
    
#define assign_byte(r) do{ \
    assign(r, new expr_num(b1)); \
    ++data; }while(0)

static void outputnegated(string s)
{
	const char *t=s.c_str(), *q=t;
	while(*t != '(')
	{
		if(!*t) { output(q); return; }
		t++;
	}
	t++;
	if(*t=='!')
		s.erase(t-q, 1);
	else
		s.insert(t-q, "!");
		
	output(s.c_str());
}

static unsigned maxpos;

static void DoJump(const byte *data, unsigned pos, unsigned &next, const char *condition=NULL)
{
	if(alreadycompiled(data, pos) || getjumpcount(data, pos) != 1)
	{
		output("%sJUMP(\33[33m%08X\33[37m)", condition?condition:"", pos);
	}
	else
	{
		if(condition && pos>next && pos<maxpos)
		{
			MarkVisited(data, pos);
			outputnegated(condition);
			RequestCompile(data, next, condition, false);
			next = pos;
		}
		else
		{
			if(condition)output(condition);
			RequestCompile(data, pos, condition);
		}
	}
}

void compile(const byte *data, unsigned pos, unsigned length, bool enterblock, unsigned endjump)
{
	unsigned oldmaxpos = maxpos;
	
	if(enterblock)
	{
		output("{");
		indentlevel += 4;
	}
	
	const byte *dd = data;
	
	data += pos;
	
	const byte *max = data+length;
	
	maxpos = pos+length;
	
	resetall();
	
	while(data < max)
	{
		pos = data-dd;
		length = max-data;
		
		MarkVisited(dd, pos);
		
		flushpendingoutput();
		dprintf("------------\n");
		
		unsigned jumpcount = getjumpcount(dd, pos);
		
		if(jumpcount)
		{
			flushpendingoutput();
			flushall(true);
			flushpendingoutput();
			printf("\33[33m%02X:%04X\33[37m:\n", pos/0x4000, ((pos>=0x4000)?0x4000:0) + (pos%0x4000));
			/*
			if(jumpcount==1)
			{
				unsigned ppo = getjump(dd, pos, 0);
				if(ppo > pos && ppo < maxpos)
				{
					output("do {}");
				}
			}
			*/
		}
		
		if(*data)showcontents();
		
		{bool j,c,e;unsigned t;
		flushpendingoutput();
		printf("// ");
		fflush(stdout);
		DisAsm(dd, pos, true,&j,&c,&e, &t);
		printf("\n");
		fflush(stdout);
		}
		
    	byte b1 = data[1];
    	byte b2 = data[2];
    	
    	unsigned short w = b1 | ((unsigned)b2 << 8);
    	
    	if(*data==0x00) ;
    	else if(*data==0x10) { ++data; output("stop(%d);", b1); }
    	
    	else if(*data==0x08) { flushpair(S,P); output("pokew(0x%04X, %s);", w, pairname(S,P)); }
    	
    	else if(*data==0x18) { flushall(); ++data; pos+=2; DoJump(dd,pos+(signed char)b1, pos);data=dd+pos-1; /* JMP disp */ }
    	else if(*data==0xC3) { flushall(); data+=2; pos+=3; DoJump(dd,w, pos); data=dd+pos-1; /* JMP abs */}
    	else if(*data==0xE9) { flushall(); output("JUMP(%s);", pairname(H,L)); /* JMP indir */}
    	else if(*data==0xC2) { flushall(); data+=2; pos+=3; DoJump(dd,w, pos, "if(!ZF)"); data=dd+pos-1; /* JMP IF NZ  */}
    	else if(*data==0xD2) { flushall(); data+=2; pos+=3; DoJump(dd,w, pos, "if(!CF)"); data=dd+pos-1; /* JMP IF CZ  */}
    	else if(*data==0xCA) { flushall(); data+=2; pos+=3; DoJump(dd,w, pos, "if(ZF)");  data=dd+pos-1; /* JMP IF ZF */}
    	else if(*data==0xDA) { flushall(); data+=2; pos+=3; DoJump(dd,w, pos, "if(CF)");  data=dd+pos-1; /* JMP IF CF */}
    	
    	else if(*data==0x20) { flushall(); ++data; pos+=2; DoJump(dd,pos+(signed char)b1, pos, "if(!ZF)"); data=dd+pos-1; /* JNZ disp */ }
    	else if(*data==0x30) { flushall(); ++data; pos+=2; DoJump(dd,pos+(signed char)b1, pos, "if(!CF)"); data=dd+pos-1; /* JNC disp */ }
    	else if(*data==0x28) { flushall(); ++data; pos+=2; DoJump(dd,pos+(signed char)b1, pos, "if(ZF)");  data=dd+pos-1; /* JZ disp */ }
    	else if(*data==0x38) { flushall(); ++data; pos+=2; DoJump(dd,pos+(signed char)b1, pos, "if(CF)");  data=dd+pos-1; /* JC disp */ }
    	
    	else if(*data==0x01)assign_word(B,C);
    	else if(*data==0x11)assign_word(D,E);
    	else if(*data==0x21)assign_word(H,L);
    	else if(*data==0x31)assign_word(S,P);
    	
    	else if(*data==0x02) { flushpair(B,C); DatAssign(new expr_pair(B,C), A); }
    	else if(*data==0x12) { flushpair(D,E); DatAssign(new expr_pair(D,E), A); }
    	else if(*data==0x22) { flushpair(H,L); output("data(%s++) = %s;", pairname(H,L), getval(A).c_str()); reset(H);reset(L); }
    	else if(*data==0x32) { flushpair(H,L); output("data(%s--) = %s;", pairname(H,L), getval(A).c_str()); reset(H);reset(L); }

    	else if(*data==0x03) { flushpair(B,C); output("++%s;", pairname(B,C)); }
    	else if(*data==0x13) { flushpair(D,E); output("++%s;", pairname(D,E)); }
    	else if(*data==0x23) { flushpair(H,L); output("++%s;", pairname(H,L)); }
    	else if(*data==0x33) { flushpair(S,P); output("++%s;", pairname(S,P)); }
    	
    	else if(*data==0x04) { assign(B, new expr_plus(B, 1)); }
    	else if(*data==0x14) { assign(D, new expr_plus(D, 1)); }
    	else if(*data==0x24) { assign(H, new expr_plus(H, 1)); }
    	else if(*data==0x34) { expr *a = new expr_ptr(new expr_pair(H,L)); output("++%s;", (a=a->calculate())->debug().c_str()); delete a; }
    	
    	else if(*data==0x05) { assign(B, new expr_minus(B, 1)); }
    	else if(*data==0x15) { assign(D, new expr_minus(D, 1)); }
    	else if(*data==0x25) { assign(H, new expr_minus(H, 1)); }
    	else if(*data==0x35) { expr *a = new expr_ptr(new expr_pair(H,L)); output("--%s;", (a=a->calculate())->debug().c_str()); delete a; }
    	
    	else if(*data==0x06) assign_byte(B);
    	else if(*data==0x16) assign_byte(D);
    	else if(*data==0x26) assign_byte(H);
    	else if(*data==0x36) { expr *a = new expr_ptr(new expr_pair(H,L)); output("%s = 0x%02X;", (a=a->calculate())->debug().c_str(), b1); ++data; delete a;}
    	
    	else if(*data==0x07) { assign(A, new expr_rcl(A)); }
    	else if(*data==0x17) { assign(A, new expr_rol(A)); }
    	else if(*data==0x27) { assign(A, new expr_daa(A)); }
    	else if(*data==0x37) { output("CF=1;"); }
    	
    	else if(*data==0x09) { flushpair(H,L);flushpair(B,C); output("%s += %s;", pairname(H,L), pairname(B,C)); reset(B);reset(C); }
    	else if(*data==0x19) { flushpair(H,L);flushpair(D,E); output("%s += %s;", pairname(H,L), pairname(D,E)); reset(D);reset(E); }
    	else if(*data==0x29) { flushpair(H,L);                output("%s += %s;", pairname(H,L), pairname(H,L)); reset(H);reset(L); }
    	else if(*data==0x39) { flushpair(H,L);flushpair(S,P); output("%s += %s;", pairname(H,L), pairname(S,P)); reset(S);reset(P); }
    	
    	else if(*data==0x0A) assign(A, new expr_ptr(new expr_pair(B,C)));
    	else if(*data==0x1A) assign(A, new expr_ptr(new expr_pair(D,E)));
    	else if(*data==0x2A) { reset(A); flushpair(H,L); output("%s = data(%s++);", regname(A), pairname(H,L)); reset(H);reset(L); }
    	else if(*data==0x3A) { reset(A); flushpair(H,L); output("%s = data(%s--);", regname(A), pairname(H,L)); reset(H);reset(L); }
    	
    	else if(*data==0x0B) { flushpair(B,C); output("--%s;", pairname(B,C)); }
    	else if(*data==0x1B) { flushpair(D,E); output("--%s;", pairname(D,E)); }
    	else if(*data==0x2B) { flushpair(H,L); output("--%s;", pairname(H,L)); }
    	else if(*data==0x3B) { flushpair(S,P); output("--%s;", pairname(S,P)); }
    	
    	else if(*data==0x0C) { assign(C, new expr_plus(C, 1)); }
    	else if(*data==0x1C) { assign(E, new expr_plus(E, 1)); }
    	else if(*data==0x2C) { assign(L, new expr_plus(L, 1)); }
    	else if(*data==0x3C) { assign(A, new expr_plus(A, 1)); }
     	
    	else if(*data==0x0D) { assign(C, new expr_minus(C, 1)); }
    	else if(*data==0x1D) { assign(E, new expr_minus(E, 1)); }
    	else if(*data==0x2D) { assign(L, new expr_minus(L, 1)); }
    	else if(*data==0x3D) { assign(A, new expr_minus(A, 1)); }
    	
    	else if(*data==0x0E) assign_byte(C);
    	else if(*data==0x1E) assign_byte(E);
    	else if(*data==0x2E) assign_byte(L);
    	else if(*data==0x3E) assign_byte(A);
    	
    	else if(*data==0x0F) { assign(A, new expr_rcr(A)); }
    	else if(*data==0x1F) { assign(A, new expr_ror(A)); }
    	else if(*data==0x2F) { assign(A, new expr_not(A)); }
    	else if(*data==0x3F) { output("CF ^= 1;"); }
    	
    	else if(*data==0x40) assign(B,B);
    	else if(*data==0x41) assign(B,C);
    	else if(*data==0x42) assign(B,D);
    	else if(*data==0x43) assign(B,E);
    	else if(*data==0x44) assign(B,H);
    	else if(*data==0x45) assign(B,L);
    	else if(*data==0x46) assign(B, new expr_ptr(new expr_pair(H,L)));
    	else if(*data==0x47) assign(B,A);
    	
    	else if(*data==0x48) assign(C,B);
    	else if(*data==0x49) assign(C,C);
    	else if(*data==0x4A) assign(C,D);
    	else if(*data==0x4B) assign(C,E);
    	else if(*data==0x4C) assign(C,H);
    	else if(*data==0x4D) assign(C,L);
    	else if(*data==0x4E) assign(C, new expr_ptr(new expr_pair(H,L)));
    	else if(*data==0x4F) assign(C,A);
    	
    	else if(*data==0x50) assign(D,B);
    	else if(*data==0x51) assign(D,C);
    	else if(*data==0x52) assign(D,D);
    	else if(*data==0x53) assign(D,E);
    	else if(*data==0x54) assign(D,H);
    	else if(*data==0x55) assign(D,L);
    	else if(*data==0x56) assign(D, new expr_ptr(new expr_pair(H,L)));
    	else if(*data==0x57) assign(D,A);
    	
    	else if(*data==0x58) assign(E,B);
    	else if(*data==0x59) assign(E,C);
    	else if(*data==0x5A) assign(E,D);
    	else if(*data==0x5B) assign(E,E);
    	else if(*data==0x5C) assign(E,H);
    	else if(*data==0x5D) assign(E,L);
    	else if(*data==0x5E) assign(E, new expr_ptr(new expr_pair(H,L)));
    	else if(*data==0x5F) assign(E,A);
    	
    	else if(*data==0x60) assign(H,B);
    	else if(*data==0x61) assign(H,C);
    	else if(*data==0x62) assign(H,D);
    	else if(*data==0x63) assign(H,E);
    	else if(*data==0x64) assign(H,H);
    	else if(*data==0x65) assign(H,L);
    	else if(*data==0x66) assign(H, new expr_ptr(new expr_pair(H,L)));
    	else if(*data==0x67) assign(H,A);
    	
    	else if(*data==0x68) assign(L,B);
    	else if(*data==0x69) assign(L,C);
    	else if(*data==0x6A) assign(L,D);
    	else if(*data==0x6B) assign(L,E);
    	else if(*data==0x6C) assign(L,H);
    	else if(*data==0x6D) assign(L,L);
    	else if(*data==0x6E) assign(L, new expr_ptr(new expr_pair(H,L)));
    	else if(*data==0x6F) assign(L,A);
    	
    	else if(*data==0x70) DatAssign(new expr_pair(H,L), B);
    	else if(*data==0x71) DatAssign(new expr_pair(H,L), C);
    	else if(*data==0x72) DatAssign(new expr_pair(H,L), D);
    	else if(*data==0x73) DatAssign(new expr_pair(H,L), E);
    	else if(*data==0x74) DatAssign(new expr_pair(H,L), H);
    	else if(*data==0x75) DatAssign(new expr_pair(H,L), L);
    	else if(*data==0x76) output("HALT(); /* wait for irq and save power */");
    	else if(*data==0x77) DatAssign(new expr_pair(H,L), A);
    	
    	else if(*data==0x78) assign(A,B);
    	else if(*data==0x79) assign(A,C);
    	else if(*data==0x7A) assign(A,D);
    	else if(*data==0x7B) assign(A,E);
    	else if(*data==0x7C) assign(A,H);
    	else if(*data==0x7D) assign(A,L);
    	else if(*data==0x7E) assign(A, new expr_ptr(new expr_pair(H,L)));
    	else if(*data==0x7F) assign(A,A);
    	
    	else if(*data==0x80) assign(A, new expr_plus(A,B));
    	else if(*data==0x81) assign(A, new expr_plus(A,C));
    	else if(*data==0x82) assign(A, new expr_plus(A,D));
    	else if(*data==0x83) assign(A, new expr_plus(A,E));
    	else if(*data==0x84) assign(A, new expr_plus(A,H));
    	else if(*data==0x85) assign(A, new expr_plus(A,L));
    	else if(*data==0x86) assign(A, new expr_plus(A, new expr_ptr(new expr_pair(H,L))));
    	else if(*data==0x87) assign(A, new expr_plus(A,A));
    	
    	else if(*data==0x88) { flushall(true); output("%s += %s + CF;", regname(A), regname(B)); }
    	else if(*data==0x89) { flushall(true); output("%s += %s + CF;", regname(A), regname(C)); }
    	else if(*data==0x8A) { flushall(true); output("%s += %s + CF;", regname(A), regname(D)); }
    	else if(*data==0x8B) { flushall(true); output("%s += %s + CF;", regname(A), regname(E)); }
    	else if(*data==0x8C) { flushall(true); output("%s += %s + CF;", regname(A), regname(H)); }
    	else if(*data==0x8D) { flushall(true); output("%s += %s + CF;", regname(A), regname(L)); }
    	else if(*data==0x8E) { flushall(true); output("%s += data(%s) + CF;", regname(A), pairname(H,L)); }
    	else if(*data==0x8F) { flushall(true); output("%s += %s + CF;", regname(A), regname(A)); }
    	
    	else if(*data==0x90) assign(A, new expr_minus(A,B));
    	else if(*data==0x91) assign(A, new expr_minus(A,C));
    	else if(*data==0x92) assign(A, new expr_minus(A,D));
    	else if(*data==0x93) assign(A, new expr_minus(A,E));
    	else if(*data==0x94) assign(A, new expr_minus(A,H));
    	else if(*data==0x95) assign(A, new expr_minus(A,L));
    	else if(*data==0x96) assign(A, new expr_minus(A, new expr_ptr(new expr_pair(H,L))));
    	else if(*data==0x97) assign(A, new expr_minus(A,A));
    	
    	else if(*data==0x98) { flushall(true); output("%s -= %s + CF;", regname(A), regname(B)); }
    	else if(*data==0x99) { flushall(true); output("%s -= %s + CF;", regname(A), regname(C)); }
    	else if(*data==0x9A) { flushall(true); output("%s -= %s + CF;", regname(A), regname(D)); }
    	else if(*data==0x9B) { flushall(true); output("%s -= %s + CF;", regname(A), regname(E)); }
    	else if(*data==0x9C) { flushall(true); output("%s -= %s + CF;", regname(A), regname(H)); }
    	else if(*data==0x9D) { flushall(true); output("%s -= %s + CF;", regname(A), regname(L)); }
    	else if(*data==0x9E) { flushall(true); output("%s -= data(%s) + CF;", regname(A), pairname(H,L)); }
    	else if(*data==0x9F) { flushall(true); output("%s -= %s + CF;", regname(A), regname(A)); }
    	
    	else if(*data==0xA0) assign(A, new expr_and(A,B));
    	else if(*data==0xA1) assign(A, new expr_and(A,C));
    	else if(*data==0xA2) assign(A, new expr_and(A,D));	
    	else if(*data==0xA3) assign(A, new expr_and(A,E));
    	else if(*data==0xA4) assign(A, new expr_and(A,H));
    	else if(*data==0xA5) assign(A, new expr_and(A,L));
    	else if(*data==0xA6) assign(A, new expr_and(A, new expr_ptr(new expr_pair(H,L))));
    	else if(*data==0xA7) assign(A, new expr_and(A,A));
    	
    	else if(*data==0xA8) assign(A, new expr_xor(A,B));
    	else if(*data==0xA9) assign(A, new expr_xor(A,C));
    	else if(*data==0xAA) assign(A, new expr_xor(A,D));
    	else if(*data==0xAB) assign(A, new expr_xor(A,E));
    	else if(*data==0xAC) assign(A, new expr_xor(A,H));
    	else if(*data==0xAD) assign(A, new expr_xor(A,L));
    	else if(*data==0xAE) assign(A, new expr_xor(A, new expr_ptr(new expr_pair(H,L))));
    	else if(*data==0xAF) assign(A, new expr_xor(A,A));
    	
    	else if(*data==0xB0) assign(A, new expr_or(A,B));
    	else if(*data==0xB1) assign(A, new expr_or(A,C));
    	else if(*data==0xB2) assign(A, new expr_or(A,D));
    	else if(*data==0xB3) assign(A, new expr_or(A,E));
    	else if(*data==0xB4) assign(A, new expr_or(A,H));
    	else if(*data==0xB5) assign(A, new expr_or(A,L));
    	else if(*data==0xB6) assign(A, new expr_or(A, new expr_ptr(new expr_pair(H,L))));
    	else if(*data==0xB7) assign(A, new expr_or(A,A));
    	
    	else if(*data==0xB8) { flush(A, true); flush(B); output("%s - %s; /* compare */", regname(A), regname(B)); }
    	else if(*data==0xB9) { flush(A, true); flush(C); output("%s - %s; /* compare */", regname(A), regname(C)); }
    	else if(*data==0xBA) { flush(A, true); flush(D); output("%s - %s; /* compare */", regname(A), regname(D)); }
    	else if(*data==0xBB) { flush(A, true); flush(E); output("%s - %s; /* compare */", regname(A), regname(E)); }
    	else if(*data==0xBC) { flush(A, true); flush(H); output("%s - %s; /* compare */", regname(A), regname(H)); }
    	else if(*data==0xBD) { flush(A, true); flush(L); output("%s - %s; /* compare */", regname(A), regname(L)); }
    	else if(*data==0xBE) { flush(A, true); flushpair(H,L); output("%s - data(%s); /* compare */", regname(A), pairname(H,L)); }
    	else if(*data==0xBF) { flush(A, true); flush(A); output("%s - %s; /* compare */", regname(A), regname(A)); }
    	
    	else if(*data==0xC0) { flushall(); output("if(!ZF)RET();"); }
    	else if(*data==0xD0) { flushall(); output("if(!CF)RET();"); }
    	else if(*data==0xE0) { ++data; DatAssign(new expr_num(0xFF00|b1), A); }
    	else if(*data==0xF0) { ++data; assign(A, new expr_ptr(0xFF00 | b1)); }
    	else if(*data==0xE2) { output("data(0xFF00+%s) = %s;", getval(C).c_str(), getval(A).c_str()); }
    	else if(*data==0xF2) assign(A, new expr_ptr(new expr_plus(C, 0xFF00)));
    	
    	else if(*data==0xC1) { flushpair(S,P); flushpair(B,C, true); /*--indentlevel;*/ output("POP(%s);", pairname(B,C)); }
    	else if(*data==0xD1) { flushpair(S,P); flushpair(D,E, true); /*--indentlevel;*/ output("POP(%s);", pairname(D,E)); }
    	else if(*data==0xE1) { flushpair(S,P); flushpair(H,L, true); /*--indentlevel;*/ output("POP(%s);", pairname(H,L)); }
    	else if(*data==0xF1) { flushpair(S,P); flush(A, true);       /*--indentlevel;*/ output("POP(AF);");                }
    	
    	
    	/* D3, E3 = nothing */
    	else if(*data==0xF3) output("disable(); /* CLI */");
    	
    	else if(*data==0xC4) { flushall(); data+=2; output("if(!ZF)CALL(0x%04X);", w); resetall(); /* CALL IF NZ  */}
    	else if(*data==0xD4) { flushall(); data+=2; output("if(!CF)CALL(0x%04X);", w); resetall(); /* CALL IF NZ  */}
    	/* E4, F4 = nothing */
    	
    	else if(*data==0xC5) { flushpair(S,P); flushpair(B,C); output("PUSH(%s);", pairname(B,C)); }
    	else if(*data==0xD5) { flushpair(S,P); flushpair(D,E); output("PUSH(%s);", pairname(D,E)); }
    	else if(*data==0xE5) { flushpair(S,P); flushpair(H,L); output("PUSH(%s);", pairname(H,L)); }
    	else if(*data==0xF5) { flushpair(S,P); flush(A);       output("PUSH(AF);");                }
    	
    	else if(*data==0xC6) { ++data; assign(A, new expr_plus(A, b1)); }
    	else if(*data==0xD6) { ++data; assign(A, new expr_minus(A, b1)); }
    	else if(*data==0xE6) { ++data; assign(A, new expr_and(A, b1)); }
    	else if(*data==0xF6) { ++data; assign(A, new expr_or(A, b1)); }
    	
    	else if(*data==0xC7) { flushall(true); output("CALL(0*8); /* RST 00 */"); }
    	else if(*data==0xD7) { flushall(true); output("CALL(2*8); /* RST 10 */"); }
    	else if(*data==0xE7) { flushall(true); output("CALL(4*8); /* RST 20 */"); }
    	else if(*data==0xF7) { flushall(true); output("CALL(6*8); /* RST 30 */"); }

    	else if(*data==0xC8) { flushall(); output("if(ZF)RET();"); }
    	else if(*data==0xD8) { flushall(); output("if(CF)RET();"); }
    	else if(*data==0xE8) { flushpair(S,P, true); ++data; output("%s += %d;", pairname(S,P), (signed char)b1); }
    	else if(*data==0xF8) { flushpair(H,L, true); flushpair(S,P); ++data; output("%s = %s + %d;", pairname(H,L), pairname(S,P), (signed char)b1); }
    	
    	else if(*data==0xC9) { flushall(true); output("RET();"); }
    	else if(*data==0xD9) { flushall(true); output("IRET();"); }
    	else if(*data==0xF9) { assign(S,H); assign(P,L); }

    	else if(*data==0xEA) { data+=2; DatAssign(new expr_num(w), A); }
    	else if(*data==0xFA) { data+=2; assign(A, new expr_ptr(w)); }
    	
    	else if(*data==0xCB)
    	{
    		++data;
    		     if(b1==0x00) assign(B, new expr_rcl(B));
    		else if(b1==0x01) assign(C, new expr_rcl(C));
    		else if(b1==0x02) assign(D, new expr_rcl(D));
    		else if(b1==0x03) assign(E, new expr_rcl(E));
    		else if(b1==0x04) assign(H, new expr_rcl(H));
    		else if(b1==0x05) assign(L, new expr_rcl(L));
    		else if(b1==0x06) { flushpair(H,L, true); output("rcl(data(%s));", pairname(H,L)); }
    		else if(b1==0x07) assign(A, new expr_rcl(A));
    		else if(b1==0x08) assign(B, new expr_rcr(B));
    		else if(b1==0x09) assign(C, new expr_rcr(C));
    		else if(b1==0x0A) assign(D, new expr_rcr(D));
    		else if(b1==0x0B) assign(E, new expr_rcr(E));
    		else if(b1==0x0C) assign(H, new expr_rcr(H));
    		else if(b1==0x0D) assign(L, new expr_rcr(L));
    		else if(b1==0x0E) { flushpair(H,L, true); output("rcr(data(%s));", pairname(H,L)); }
    		else if(b1==0x0F) assign(A, new expr_rcr(A));
    		else if(b1==0x10) assign(B, new expr_rol(B));
    		else if(b1==0x11) assign(C, new expr_rol(C));
    		else if(b1==0x12) assign(D, new expr_rol(D));
    		else if(b1==0x13) assign(E, new expr_rol(E));
    		else if(b1==0x14) assign(H, new expr_rol(H));
    		else if(b1==0x15) assign(L, new expr_rol(L));
    		else if(b1==0x16) { flushpair(H,L, true); output("rol(data(%s));", pairname(H,L)); }
    		else if(b1==0x17) assign(A, new expr_rol(A));
    		else if(b1==0x18) assign(B, new expr_ror(B));
    		else if(b1==0x19) assign(C, new expr_ror(C));
    		else if(b1==0x1A) assign(D, new expr_ror(D));
    		else if(b1==0x1B) assign(E, new expr_ror(E));
    		else if(b1==0x1C) assign(H, new expr_ror(H));
    		else if(b1==0x1D) assign(L, new expr_ror(L));
    		else if(b1==0x1E) { flushpair(H,L, true); output("ror(data(%s));", pairname(H,L)); }
    		else if(b1==0x1F) assign(A, new expr_ror(A));
    		else if(b1==0x20) assign(B, new expr_shl(B, 1));
    		else if(b1==0x21) assign(C, new expr_shl(C, 1));
    		else if(b1==0x22) assign(D, new expr_shl(D, 1));
    		else if(b1==0x23) assign(E, new expr_shl(E, 1));
    		else if(b1==0x24) assign(H, new expr_shl(H, 1));
    		else if(b1==0x25) assign(L, new expr_shl(L, 1));
    		else if(b1==0x26) { flushpair(H,L, true); output("data(%s) <<= 1;", pairname(H,L)); }
    		else if(b1==0x27) assign(A, new expr_shl(A, 1));
    		else if(b1==0x28) assign(B, new expr_sar(B, 1));
    		else if(b1==0x29) assign(C, new expr_sar(C, 1));
    		else if(b1==0x2A) assign(D, new expr_sar(D, 1));
    		else if(b1==0x2B) assign(E, new expr_sar(E, 1));
    		else if(b1==0x2C) assign(H, new expr_sar(H, 1));
    		else if(b1==0x2D) assign(L, new expr_sar(L, 1));
    		else if(b1==0x2E) { flushpair(H,L, true); output("(signed char)data(%s) >>= 1;", pairname(H,L)); }
    		else if(b1==0x2F) assign(A, new expr_sar(A, 1));
    		else if(b1==0x30) assign(B, new expr_swap(B));
    		else if(b1==0x31) assign(C, new expr_swap(C));
    		else if(b1==0x32) assign(D, new expr_swap(D));
    		else if(b1==0x33) assign(E, new expr_swap(E));
    		else if(b1==0x34) assign(H, new expr_swap(H));
    		else if(b1==0x35) assign(L, new expr_swap(L));
    		else if(b1==0x36) { flushpair(H,L, true); output("ror(data(%s), 4);", pairname(H,L)); }
    		else if(b1==0x37) assign(A, new expr_swap(A));
    		else if(b1==0x38) assign(B, new expr_shr(B, 1));
    		else if(b1==0x39) assign(C, new expr_shr(C, 1));
    		else if(b1==0x3A) assign(D, new expr_shr(D, 1));
    		else if(b1==0x3B) assign(E, new expr_shr(E, 1));
    		else if(b1==0x3C) assign(H, new expr_shr(H, 1));
    		else if(b1==0x3D) assign(L, new expr_shr(L, 1));
    		else if(b1==0x3E) { flushpair(H,L, true); output("data(%s) >>= 1;", pairname(H,L)); }
    		else if(b1==0x3F) assign(A, new expr_shr(A, 1));
    		else
    		{
    			expr **tm = NULL;
    			switch(b1&7)
    			{
    				case 0: tm = &B; break;
    				case 1: tm = &C; break;
    				case 2: tm = &D; break;
    				case 3: tm = &E; break;
    				case 4: tm = &H; break;
    				case 5: tm = &L; break;
    				case 7: tm = &A; break;
    			}
    			
    			expr *x = new expr_ptr(new expr_pair(H,L));
    			
    			if(!tm)
    			{
    				x = x->calculate();
    				register byte c = ((1+~b1) >> 3) & 7;
    				flushpair(H,L);
    				switch(b1 >> 6) // 0x40, 0x80, 0xC0
    				{
    					case 1: output("test(%s, 0x%02X);", x->debug().c_str(), 1<<c); break;
    					case 2: output("and(%s, 0x%02X);", x->debug().c_str(), (1<<c)^255); break;
        				case 3: output("or(%s, 0x%02X);", x->debug().c_str(), 1<<c); break;
        			}
        			delete x;
        		}
    			else
    			{
    				switch(b1 & 0xF8)
    				{
        				case 0x40: x = (new expr_test(*tm, 1<<0))->calculate(); output("%s;", x->debug().c_str()); delete x; break;
        				case 0x78: x = (new expr_test(*tm, 1<<1))->calculate(); output("%s;", x->debug().c_str()); delete x; break;
        				case 0xB0: x = (new expr_test(*tm, 1<<2))->calculate(); output("%s;", x->debug().c_str()); delete x; break;
        				case 0xE8: x = (new expr_test(*tm, 1<<3))->calculate(); output("%s;", x->debug().c_str()); delete x; break;
        				case 0x60: x = (new expr_test(*tm, 1<<4))->calculate(); output("%s;", x->debug().c_str()); delete x; break;
        				case 0x98: x = (new expr_test(*tm, 1<<5))->calculate(); output("%s;", x->debug().c_str()); delete x; break;
        				case 0xD0: x = (new expr_test(*tm, 1<<6))->calculate(); output("%s;", x->debug().c_str()); delete x; break;
        				case 0x48: x = (new expr_test(*tm, 1<<7))->calculate(); output("%s;", x->debug().c_str()); delete x; break;
        				
        				case 0x80: assign(*tm, new expr_and(*tm, (1<<0)^255)); break;     
        				case 0xB8: assign(*tm, new expr_and(*tm, (1<<1)^255)); break;     
        				case 0xF0: assign(*tm, new expr_and(*tm, (1<<2)^255)); break;     
        				case 0x68: assign(*tm, new expr_and(*tm, (1<<3)^255)); break;     
        				case 0xA0: assign(*tm, new expr_and(*tm, (1<<4)^255)); break;     
        				case 0xD8: assign(*tm, new expr_and(*tm, (1<<5)^255)); break;     
        				case 0x50: assign(*tm, new expr_and(*tm, (1<<6)^255)); break;     
        				case 0x88: assign(*tm, new expr_and(*tm, (1<<7)^255)); break;     
        				                                                       
        				case 0xC0: assign(*tm, new expr_or(*tm, 1<<0)); break;            
        				case 0xF8: assign(*tm, new expr_or(*tm, 1<<1)); break;            
        				case 0x70: assign(*tm, new expr_or(*tm, 1<<2)); break;            
        				case 0xA8: assign(*tm, new expr_or(*tm, 1<<3)); break;            
        				case 0xE0: assign(*tm, new expr_or(*tm, 1<<4)); break;            
        				case 0x58: assign(*tm, new expr_or(*tm, 1<<5)); break;            
        				case 0x90: assign(*tm, new expr_or(*tm, 1<<6)); break;            
        				case 0xC8: assign(*tm, new expr_or(*tm, 1<<7)); break;            
        			}
        		}
    		}                                                              
    	}
    	                                                                   
    	/* DB, EB = nothing */
    	else if(*data==0xFB) { output("enable(); /* STI */"); }
    	
    	else if(*data==0xCC) { flushall(true); data+=2; output("if(ZF)CALL(0x%04X);", w); /* CALL IF ZF  */ }
    	else if(*data==0xDC) { flushall(true); data+=2; output("if(CF)CALL(0x%04X);", w); /* CALL IF CF  */ }
    	/* EC, FC = nothing */
    	
    	else if(*data==0xCD) { flushall(true); data+=2; output("CALL(0x%04X);", w); }
    	/* DD, ED, FD = nothing */
    	
    	else if(*data==0xCE) { flush(A, true); output("%s += 0x%02X + CF;", regname(A), b1); }
    	else if(*data==0xDE) { flush(A, true); output("%s -= 0x%02X + CF;", regname(A), b1); }
    	else if(*data==0xEE) { ++data; assign(A, new expr_xor(A, b1)); }
    	else if(*data==0xFE) { ++data; output("%s - 0x%02X; /* compare */", getval(A).c_str(), b1); }
    	
    	else if(*data==0xCF) { flushall(true); output("CALL(1*8); /* RST 08 */"); }
    	else if(*data==0xDF) { flushall(true); output("CALL(3*8); /* RST 18 */"); }
    	else if(*data==0xEF) { flushall(true); output("CALL(5*8); /* RST 28 */"); }
    	else if(*data==0xFF) { flushall(true); output("CALL(7*8); /* RST 38 */"); }
    	else
    		output("/* unknown opcode: 0x%02X */", *data);
    		
    	++data;
    }
    		
    flushall();
    
    if(endjump)
    	DoJump(dd, endjump, pos);
    
	if(enterblock)
	{
		indentlevel -= 4;
		output("}");
	}

   	flushpendingoutput();
   	
   	maxpos = oldmaxpos;
}
