#ifndef __DASM_EXPR_H
#define __DASM_EXPR_H

#ifndef MAKEDEPEND
#include <string>
#include <cstdio>
#endif

extern const string checkpair(class expr *l, class expr *r);

using namespace std;

class expr
{
public:
	int refnum;
	bool isconst;
	int regid;
	string varname;
	bool notneedflush;
	
	expr() :                     refnum(0), isconst(false), regid(0), varname("?"), notneedflush(false) { }
	expr(const string v, int r): refnum(0), isconst(false), regid(r), varname(v), notneedflush(false)   { }
	virtual ~expr()
	{
		extern void flushpendingoutput();
		flushpendingoutput();
//		printf("%s (at %X) is erasing (regid %d).\n", this->debug().c_str(), (unsigned)this, regid);
	}
	
	virtual const string debug() const
	{
		return varname;
	}	
	virtual unsigned getnum() const
	{
		return 0;
	}
	virtual bool depends(int r) const
	{
		return r && r==regid;
	}
	virtual bool depends(const string &s) const
	{
		return debug() == s;
	}
	virtual expr *calculate()
	{
		// if it needs to generate other type of
		// expr, it deletes itself and returns new.
		return this; 
	}
};

class expr_num : public expr
{
	unsigned n;
public:
	expr_num(unsigned num) : n(num)
	{
		isconst = true;
	}
	virtual const string debug() const
	{
		char Buf[32];
		if(n < 16)
			sprintf(Buf, "%u", n);
		else
		{
		/*	unsigned t = n;
			while(!(t&1))t>>=1;
			if(t==1)
				sprintf(Buf, "%u", n);
			else
		*/		sprintf(Buf, "0x%X", n);
		}
		return Buf;
	}
	virtual unsigned getnum() const
	{
		return n;
	}
};

class expr_2op : public expr
{
	expr *l, *r;
	string op;
	unsigned (*fun)(unsigned, unsigned);
public:
	expr_2op(expr *left, expr *right, const char *o, unsigned (*f)(unsigned,unsigned)) : op(o), fun(f)
	{
		l = left->calculate();
		r = right->calculate();
		++left->refnum;
		++right->refnum;
	}
	expr_2op(expr *left, unsigned right, const char *o, unsigned (*f)(unsigned,unsigned)) : op(o), fun(f)
	{
		l = left->calculate();
		r = new expr_num(right);
		++l->refnum;
		++r->refnum;
	}
	void release()
	{
		if(l && !--l->refnum) { delete l; l=NULL; }
		if(r && !--r->refnum) { delete r; r=NULL; }
	}
	virtual ~expr_2op()
	{
		release();
	}
	virtual bool depends(int rid) const
	{
		return l->depends(rid) || r->depends(rid);
	}
	virtual bool depends(const string &s) const
	{
		return debug() == s || l->depends(s) || r->depends(s);
	}
	virtual const string debug() const
	{
		return '(' + l->debug() + ' ' + op + ' ' + r->debug() + ')';
	}
	virtual expr *calculate()
	{
		if(((op=="xor"||op=="-") && l==r)
		 || (op=="and" && r->isconst && !r->getnum()))
		{
			if(!refnum)delete this;else release();
			return new expr_num(0);
		}
		if(op=="and" && r==l)
		{
			expr *tmp = l;
			l = NULL;
			if(!refnum)delete this;else release();
			
			return tmp;
		}
		if(l->isconst && r->isconst)
		{
			--l->refnum; l = l->calculate(); ++l->refnum;
			--r->refnum; r = r->calculate(); ++r->refnum;
			unsigned n = fun(l->getnum(), r->getnum());
			if(!refnum)delete this;else release();
			return new expr_num(n);
		}
		return this;
	}
};

#define def_2op_class(name, act, char) \
	static unsigned expr_fun_##name(unsigned a, unsigned b) { return act; } \
	class expr_##name : public expr_2op \
	{ \
	public: \
		expr_##name(expr *left, expr *right) : expr_2op(left, right, char, expr_fun_##name) { } \
		expr_##name(expr *left, unsigned right) : expr_2op(left, right, char, expr_fun_##name) { } \
	}

def_2op_class(plus,  a+b, "+");
def_2op_class(minus, a-b, "-");
def_2op_class(xor,   a^b, "xor");
def_2op_class(test,  a^b, "test");
def_2op_class(and,   a&b, "and");
def_2op_class(or,    a|b, "or");
def_2op_class(shl,  a<<b, "shl");
def_2op_class(shr,  a>>b, "shr");

class expr_1own : public expr
{
protected:
	expr *i;
public:
	void release()
	{
		if(i && !--i->refnum) { delete i; i=NULL; }
	}
	virtual ~expr_1own()
	{
		release();
	}
	virtual bool depends(int rid) const
	{
		return i->depends(rid);
	}
	virtual bool depends(const string &s) const
	{
		return debug() == s || i->depends(s);
	}
};

class expr_swap : public expr_1own
{
public:
	expr_swap(expr *item)
	{
		i = item->calculate();
		++i->refnum;
	}
	virtual const string debug() const
	{
		return "swap(" + i->debug() + ')';
	}
	virtual expr *calculate()
	{
		if(i->isconst)
		{
			--i->refnum; i = i->calculate(); ++i->refnum;
			unsigned n = i->getnum();
			if(!refnum)delete this;else release();
			return new expr_num(((n&15) << 4) | (n>>4));
		}
		return this;
	}
};

class expr_nonjob : public expr_1own
{
	const char *fname;
public:
	expr_nonjob(expr *item, const char *fn) : fname(fn)
	{
		i = item->calculate();
		++i->refnum;
	}
	virtual const string debug() const
	{
		return fname + i->debug() + ')';
	}
};

/* rol and rcl are swapped, that's correct. */
class expr_rcl : public expr_nonjob { public: expr_rcl(expr *item) : expr_nonjob(item, "ROL(") { }; };
class expr_rol : public expr_nonjob { public: expr_rol(expr *item) : expr_nonjob(item, "RCL(") { }; };
class expr_rcr : public expr_nonjob { public: expr_rcr(expr *item) : expr_nonjob(item, "ROR(") { }; };
class expr_ror : public expr_nonjob { public: expr_ror(expr *item) : expr_nonjob(item, "RCR(") { }; };
class expr_daa : public expr_nonjob { public: expr_daa(expr *item) : expr_nonjob(item, "DAA(") { }; };

class expr_pair : public expr
{
	expr *h, *l;
public:
	expr_pair(expr *hi, expr *lo)
	{
		h = hi->calculate();
		l = lo->calculate();
		++h->refnum;
		++l->refnum;
	}
	void release()
	{
		if(h && !--h->refnum) { delete h; h=NULL; }
		if(l && !--l->refnum) { delete l; l=NULL; }
	}	
	virtual ~expr_pair()
	{
		release();
	}
	virtual const string debug() const
	{
		string s = checkpair(h, l);
		if(s.size())return s;
		return "pair(" + h->debug() + ',' + l->debug() + ')';
	}
	virtual bool depends(int rid) const
	{
		return h->depends(rid) || l->depends(rid);
	}
	virtual expr *calculate()
	{
		--h->refnum; h = h->calculate(); ++h->refnum;
		--l->refnum; l = l->calculate(); ++l->refnum;
		if(h->isconst && l->isconst)
		{
			unsigned n = (h->getnum() << 8) | l->getnum();
			if(!refnum)delete this;else release();
			return new expr_num(n);
		}
		return this;
	}
};

class expr_ptr : public expr_1own
{
public:
	expr_ptr(expr *item)
	{
		i = item->calculate();
		++i->refnum;
	}
	expr_ptr(unsigned ptr)
	{
		i = new expr_num(ptr);
		++i->refnum;
	}
	virtual const string debug() const
	{
		return "data(" + i->debug() + ')';
	}
	virtual expr *calculate()
	{
		--i->refnum; i = i->calculate(); ++i->refnum;
		return this;
	}
};

class expr_sar : public expr_1own
{
	unsigned n;
public:
	expr_sar(expr *item, unsigned num) : n(num)
	{
		i = item->calculate();
		++i->refnum;
	}
	virtual const string debug() const
	{
		char Buf[32];
		sprintf(Buf, " sar %u)", n);
		return '(' + i->debug() + Buf;
	}
	virtual expr *calculate()
	{
		if(i->isconst)
		{
			--i->refnum; i = i->calculate(); ++i->refnum;
			unsigned v = i->getnum(), c=n;
			if(!refnum)delete this;else release();
			
			return new expr_num((signed char)v >> c);
		}
		return this;
	}
};

class expr_not : public expr_1own
{
public:
	expr_not(expr *item)
	{
		i = item->calculate();
		++i->refnum;
	}
	virtual bool depends(int rid) const
	{
		return i->depends(rid);
	}
	virtual const string debug() const
	{
		return "(not " + i->debug() + ')';
	}
	virtual expr *calculate()
	{
		if(i->isconst)
		{
			--i->refnum; i = i->calculate(); ++i->refnum;
			unsigned v = i->getnum();
			if(!refnum)delete this;else release();
			return new expr_num(v ^ 0xFF);
		}
		return this;
	}
};

#undef def_2op_class

#endif
